from .average_recommender import AverageRecommender
from .random_recommender import RandomRecommender
from .knn_meta_recommender import KNNMetaRecommender
from .surprise_recommenders import (
    CoClusteringRecommender,
    KNNWithMeansRecommender,
    KNNDatasetRecommender,
    KNNMLRecommender,
    SlopeOneRecommender,
    SVDRecommender)
